"""Ferda public request API."""
from typing import Dict, List, cast

from asgiref.sync import sync_to_async
from django.http import HttpRequest as Request
from ferda.api import Error, check_permissions
from ferda.constants import LoggerResult
from ninja import Router
from prequel import PublicRequest, PublicRequestId, RegistryObjectType
from prequel.exceptions import CannotBeProcessed, PublicRequestDoesNotExist

from .backend import PUBREQ_INFO, PUBREQ_PROCESSING
from .constants import LogEntryType, LogReferenceType
from .utils import create_log_entry

router = Router(tags=["public-request"])

# URLs
# api/public-request : base
#                   /-/get-types/ : get_types
#                   /{public_request_id}/ : get
#                   /?type={type}&value={value} : list
#                   /{public_request_id}/resolve/ : resolve
#                   /{public_request_id}/invalidate/ : invalidate


@router.get("/-/get-types/", response={200: Dict[str, bool]}, summary="Public request types")
async def get_types(request: Request):  # type: ignore[no-untyped-def]
    """Get list of supported public request types with their staff resolvable flag."""
    # no need for logging here
    return await PUBREQ_INFO.get_types()


@router.get(
    "/{public_request_id}/",
    response={200: PublicRequest, 404: Error},
    summary="Public request details",
)
async def get(  # type: ignore[no-untyped-def]
    request: Request, public_request_id: PublicRequestId
):
    """Get details about public request (given by identifier)."""
    await sync_to_async(check_permissions)(request, permissions=["ferda_pubreq.can_view"])

    log_entry = await create_log_entry(
        request,
        LogEntryType.PUBLIC_REQUEST_GET,
        properties={"public_request_id": public_request_id},
    )
    with log_entry:
        try:
            response = await PUBREQ_INFO.get(PublicRequestId(public_request_id))
            log_entry.result = LoggerResult.SUCCESS
            log_reference_type = LogReferenceType(
                RegistryObjectType(response.reference.type).value
            )
            log_entry.references = {log_reference_type: response.reference.value}
            return response
        except PublicRequestDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail="Public request not found")


@router.get("/", response={200: List[PublicRequest]}, summary="List public requests")
async def list(  # type: ignore[no-untyped-def]
    request: Request, ref_type: RegistryObjectType, ref_value: str
):
    """List all public requests linked to given object.

    Supported object types: contact, domain, nsset, keyset
    """
    await sync_to_async(check_permissions)(request, permissions=["ferda_pubreq.can_view"])

    log_entry = await create_log_entry(
        request,
        LogEntryType.PUBLIC_REQUEST_LIST,
        properties={"ref_type": ref_type.value, "ref_value": ref_value},
    )
    with log_entry:
        response = await PUBREQ_INFO.list(ref_type.value, ref_value)
        log_entry.result = LoggerResult.SUCCESS
        if response:  # more info for non-empty list
            log_entry.properties = {"items_count": str(len(response))}
            log_entry.references = {LogReferenceType(ref_type.value): ref_value}
        return response


@router.put(
    "/{public_request_id}/resolve/",
    response={200: None, 404: Error, 409: Error},
    summary="Resolve public request.",
)
async def resolve(  # type: ignore[no-untyped-def]
    request: Request, public_request_id: PublicRequestId
):
    """Resolve public request (given by identifier)."""
    await sync_to_async(check_permissions)(request, permissions=["ferda_pubreq.can_process"])

    log_entry = await create_log_entry(
        request,
        LogEntryType.PUBLIC_REQUEST_RESOLVE,
        properties={"public_request_id": public_request_id},
    )
    with log_entry:
        try:
            await PUBREQ_PROCESSING.resolve(
                public_request_id, cast(str, log_entry.entry_id)
            )  # entry_id is sure to be set to string here
            log_entry.result = LoggerResult.SUCCESS
            log_entry.references = {LogReferenceType.PUBLIC_REQUEST: public_request_id}
        except PublicRequestDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail="Public request not found")
        except CannotBeProcessed:
            log_entry.result = LoggerResult.FAIL
            log_entry.references = {LogReferenceType.PUBLIC_REQUEST: public_request_id}
            return 409, Error(detail="Public request cannot be processed")


@router.put(
    "/{public_request_id}/invalidate/",
    response={200: None, 404: Error, 409: Error},
    summary="Invalidate public request.",
)
async def invalidate(  # type: ignore[no-untyped-def]
    request: Request, public_request_id: PublicRequestId
):
    """Invalidate public request (given by identifier)."""
    await sync_to_async(check_permissions)(request, permissions=["ferda_pubreq.can_process"])

    log_entry = await create_log_entry(
        request,
        LogEntryType.PUBLIC_REQUEST_INVALIDATE,
        properties={"public_request_id": public_request_id},
    )
    with log_entry:
        try:
            await PUBREQ_PROCESSING.invalidate(
                public_request_id, cast(str, log_entry.entry_id)
            )  # entry_id is sure to be set to string here
            log_entry.result = LoggerResult.SUCCESS
            log_entry.references = {LogReferenceType.PUBLIC_REQUEST: public_request_id}
        except PublicRequestDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail="Public request not found")
        except CannotBeProcessed:
            log_entry.result = LoggerResult.FAIL
            log_entry.references = {LogReferenceType.PUBLIC_REQUEST: public_request_id}
            return 409, Error(detail="Public request cannot be processed")
